#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string>
#include <stdio.h>
#include <ctype.h>
#include <iostream>


using namespace std;


#define NUM_TYPES 6 
string tinyTypes[NUM_TYPES] = {"sbyte", "sword", "slong", "ubyte", "uword", "ulong"};

#define NUM_OPERS 11 
string tinyOpers[NUM_OPERS] = {"+", "-", "*", "/",	// Arithmetic
                               "&", "|", "~",		// Bitwize logical
			       "=", "<>", "<=", ">="};	// Relational

string cTypes[NUM_TYPES] = {"sbyte", "sword", "slong", "ubyte", "uword", "ulong"};
string cOpers[NUM_OPERS] = {"+", "-", "*", "/",		// Arithmetic
                             "&", "|", "^",		// Bitwize logical
			    "==", "!=", "<=", ">="};	// Relational



string* types;
string* opers;

string language;

int testNum = 0;

void genIncludes()
{
	cout << "#include <unistd.h>"    << endl;
        cout << "#include <sys/types.h>" << endl;
        cout << "#include <sys/stat.h>"  << endl;
        cout << "#include <fcntl.h>"     << endl;
        cout << "#include <stdio.h>"     << endl;
        cout << "#include <ctype.h>"     << endl;
        cout << "#include \"iolib.h\""   << endl;
	cout << endl;
}


void genDefines()
{
	cout << "#define sbyte signed char\n" << endl;
	cout << "#define sword signed short\n" << endl;
	cout << "#define slong signed int\n" << endl;
	cout << "#define ubyte unsigned char\n" << endl;
	cout << "#define uword unsigned short\n" << endl;
	cout << "#define ulong unsigned int\n" << endl;
}


void genDataDecls()
{
	cout << "sbyte sbyteRes = 0;" << endl;
	cout << "sword swordRes = 0;" << endl;
	cout << "slong slongRes = 0;" << endl;
	cout << "ubyte ubyteRes = 0;" << endl;
	cout << "uword uwordRes = 0;" << endl;
	cout << "ulong ulongRes = 0;" << endl;

	cout << "sbyte sbyte001 = 43210;" << endl;
	cout << "sword sword001 = 43211;" << endl;
	cout << "slong slong001 = 43210;" << endl;
	cout << "ubyte ubyte001 = 43211;" << endl;
	cout << "uword uword001 = 43210;" << endl;
	cout << "ulong ulong001 = 43210;" << endl;

	cout << "sbyte sbyte002 = 43211;" << endl;
	cout << "sword sword002 = 43210;" << endl;
	cout << "slong slong002 = 43211;" << endl;
	cout << "ubyte ubyte002 = 43210;" << endl;
	cout << "uword uword002 = 43211;" << endl;
	cout << "ulong ulong002 = 43210;" << endl;
	cout << endl;
}

void genTests()
{
	int operIdx;
	int resultTypeIdx;
	int arg01TypeIdx;
	int arg02TypeIdx;

	string oper;
	string resultType;
	string arg01Type;
	string arg02Type;
	
	
	for (operIdx = 0; operIdx < NUM_OPERS; operIdx++)
	{
		oper = opers[operIdx];
		for (resultTypeIdx = 0; resultTypeIdx < NUM_TYPES; resultTypeIdx++)
		{
			resultType = types[resultTypeIdx];
			for (arg01TypeIdx = 0; arg01TypeIdx < NUM_TYPES; arg01TypeIdx++)
			{
				arg01Type = types[arg01TypeIdx];
				for (arg02TypeIdx = 0; arg02TypeIdx < NUM_TYPES; arg02TypeIdx++)
				{
					arg02Type = types[arg02TypeIdx];
					
					cout << "\t";
					cout << resultType + "Res";
					cout << " = ";
					cout << arg01Type + "001 ";
					cout << oper;
					cout << " " << arg02Type + "002";
					cout << ";\n";
					
					if (language == "c")
					{
						cout << "\twrite" << resultType <<
					            "(" << resultType << "Res);" << endl;
					}
					else
					{
						cout << "\twrite(" <<
						     resultType << "Res);" << endl;
					}
				}
			}
		}
	}
}


void genProlog()
{
	if (language == "c")
	{
		cout << "int main (int argc, char* argv[])" << endl;
		cout << "{" << endl;
	}
	else
	{
		cout << "begin" << endl; 
	}
}


void genEpilog()
{
	if (language == "c")
	{
		cout << "\nreturn(0);" << endl;
		cout << "}" << endl;
	}
	else
	{
		cout << "end." << endl; 
	}
}


void usage()
{
	cerr << "Usage:" << endl;
	cerr << "\topertestgen language" << endl;
}


void languageSetup()
{
	if (language == "tiny")
	{
		types = tinyTypes;
		opers = tinyOpers;
	}
	else
	{
		types = cTypes;
		opers = cOpers;
	}	
}


int main (int argc, char* argv[])
{
	if (argc == 2)
	{
		language = string(argv[1]);
		if ((language != "c") && (language != "tiny"))
		{
			cerr << "Unknown language " << language << endl; 
			exit (1);
		}
	}
	else
	{
		usage();
		exit (1);
	}

	cerr << "language = " << language << endl;

	languageSetup();

	if (language == "c")
	{
		genIncludes();

		genDefines ();
	}

	genDataDecls();
	
	genProlog(); 

	genTests();

	genEpilog();

	return (0);
}
