// output_68000.cpp

#include "global.h"
#include "output.h"

unsigned int LCount;	// Lable counter

// Output a String with Tab
void Emit(string s)
{
	cout << "\t" << s;
}


void WriteLn(string s)
{
	cout << s << endl;
}


// Output a String with Tab and CRLF
void EmitLn(string s)
{
	Emit(s);
	WriteLn("");
}


// Generate a Unique Label
string* NewLabel()
{
	char buff[10];
	string* sl = new string("L");
	sprintf (buff, "%d", LCount);
	string sn((char*)buff);
        LCount++;
	*sl = *sl + sn;
	return (sl);
}


// Post a Label To Output
void PostLabel(string L)
{
	WriteLn(L + ":");
}


// Clear the primary register
void Clear()
{
	EmitLn("CLR D0");
}


// Negate the primary register
void Negate()
{
	EmitLn("NEG D0");
}


// Complement the primary register 
void NotIt()
{
	EmitLn("NOT D0");
}


// Load a constant into the primary register
void LoadConst(string c)
{
	cout << "\tMOVE #" << c << ",D0" << endl;
}


// Load a variable into the primary register
void LoadVar(string Name)
{
	if (!InTable(Name))
	{
		Undefined (Name);
	}
	EmitLn("MOVE " + Name + "(PC),D0");
}


// Push primary register on to stack
void Push()
{
	EmitLn("MOVE D0,-(SP)");
}
  

// Add top of stack to primary
void PopAdd()
{
	EmitLn("ADD (SP)+,D0");
}


// Subtract primary from top of stack
void PopSub()
{
	EmitLn("SUB (SP)+,D0");
	EmitLn("NEG D0");
}


// Multiply top of stack by primary
void PopMul()
{
	EmitLn("MULS (SP)+,D0");
}


// Divide top of stack by primary
void PopDiv()
{
	EmitLn("MOVE (SP)+,D7");
	EmitLn("EXT.L D7");
	EmitLn("DIVS D0,D7");
	EmitLn("MOV D7,D0");
}


// AND top of stack with primary register
void PopAnd()
{
	EmitLn("AND (SP)+,D0");
}


// OR top of stack with primary register
void PopOr()
{
	EmitLn("OR (SP)+,D0");
}


// XOR top of stack with primary register
void PopXor()
{
	EmitLn("XOR (SP)+,D0");
}


// Compare top of stack with primary register
void PopCompare ()
{
	EmitLn("CMP (SP)+,D0");
}


// Set D0 if compare was = 
void SetEqual ()
{
	EmitLn("SEQ D0");
	EmitLn("EXT D0");
}


// Set D0 if compare was != 
void SetNotEqual()
{
	EmitLn("SNE D0");
	EmitLn("EXT D0");
}


// Set D0 if compare was > 
void SetGreater()
{
	EmitLn("SLT D0");
	EmitLn("EXT D0");
}


// Set D0 if compare was < 
void SetLess()
{
	EmitLn("SGT D0");
	EmitLn("EXT D0");
}


// Set D0 if compare was <= 
void SetLessOrEqual()
{
	EmitLn("SGE D0");
	EmitLn("EXT D0");
}


// Set D0 if compare was >= 
void SetGreaterOrEqual()
{
	EmitLn("SLE D0");
	EmitLn("EXT D0");
}


// Store primary to variable
void Store (string Name)
{
	EmitLn("LEA " + Name + "(PC),A0");
	EmitLn("MOVE D0,(A0)");
}


// Branch unconditional 
void Branch(string l)
{
	EmitLn("BRA " + l);
}


// Branch False 
void BranchFalse(string l)
{
	EmitLn("TST D0");
	EmitLn("BEQ " + l);
}


// Read variable into primary register
void ReadIt(string Name)
{
	EmitLn ("BSR READ");
	Store(Name);
}


// Write variable from primary register
void WriteIt()
{
	EmitLn("BSR WRITE");
}


// Write header info
void Header()
{
	WriteLn ("WARMST\tEQU $A01E");
	EmitLn ("LIB TINYLIB");
}


// Write the prolog
void Prolog()
{
	PostLabel ("MAIN");
}


// Write the epilog
void Epilog()
{
	EmitLn("DC WARMST");
	EmitLn("END MAIN");
}


// Allocate storage for variable
void Allocate (string Name, string Val)
{
	WriteLn(Name + ":\tDC " + Val);
}


// Call a procedure
void Call(string Name)
{
     EmitLn("BSR " + Name);
}


// Return from subroutine
void Return()
{
	EmitLn("RTS");
}


// Load a parameter into the primary register
void LoadParam(int n)
{
	int offset;
	char offsetStr[100];
	offset = 8 + 2*(NumParams - n);
	sprintf(offsetStr, "%d", offset);
	EmitLn (string("MOVE ") + offsetStr + "(A6),D0");
}


// Store a parameter from the primary register
void StoreParam(int n)
{
	int offset;
	char offsetStr[100];
	offset = 8 + 2*(NumParams - n);
	sprintf(offsetStr, "%d", offset);
	EmitLn (string("MOVE D0,") + offsetStr + "(A6)");
}

// Adjust the stack pointer upwards by N bytes
void CleanStack(int N)
{
	char Nstr[100];
	sprintf (Nstr, "%d", N);
	if (N > 0)
	{
		EmitLn (string ("ADD #") + Nstr + ",SP");
	}
}


// Write the prolog for a procedure
void ProcProlog(string N)
{
	PostLabel(N);
	EmitLn("LINK A6,#0");
}


// Write the epilog for a procedure
void ProcEpilog()
{
	EmitLn("UNLK A6");
	EmitLn("RTS");
}
