// output_x86.cpp

#include "global.h"
#include "output.h"
#include "symtab.h"

unsigned int LCount;	// Lable counter


string Str (unsigned long long N)
{
	char S[20];
	sprintf (S, "%u", N);
	return (string(S));
}


// Output a String with Tab
void Emit(string s)
{
	cout << "\t" << s;
}


void WriteLn(string s)
{
	cout << s << endl;
}


// Output a String with Tab and CRLF
void EmitLn(string s)
{
	Emit(s);
	WriteLn("");
}


string TypeToStr(int Type)
{
	switch (Type)
	{
		case TYPE_BYTE:
			return string("sbyte");
			break;
		case TYPE_UBYTE:
			return string("ubyte");
			break;
		case TYPE_WORD:
			return string ("sword");
			break;
		case TYPE_UWORD:
			return string ("uword");
			break;
		case TYPE_LONG:
			return string ("slong");
			break;
		case TYPE_ULONG:
			return string ("ulong");
			break;
		default:
			return string ("unknown");
			break;
	}
}


void PrintTypes (int Type1, int Type2, int Type)
{
/*
	cout << "# " << TypeToStr(Type1) <<
		", " << TypeToStr(Type2) << 
		" becomes " << TypeToStr(Type) << endl;
*/
}




// Generate a Unique Label
string* NewLabel()
{
	char buff[10];
	string* sl = new string(".L");
	sprintf (buff, "%d", LCount);
	string sn((char*)buff);
        LCount++;
	*sl = *sl + sn;
	return (sl);
}


// Post a Label To Output
void PostLabel(string L)
{
	WriteLn(L + ":");
}


// Clear the primary register
int Clear()
{
	EmitLn("clr\t%eax");
	return (TYPE_LONG);	
}


// Negate the primary register
void Negate()
{
	EmitLn("neg\t%eax");
}


// Complement the primary register 
int NotIt(int Type)
{
	EmitLn("not\t%eax");
	return (Type);
}


// Load a constant into the primary register
void LoadConst(string c, int Type)
{
	// XXX Type not used as we always use longs internally
	EmitLn ("movl\t$" + c + ", %eax");
}


// Load a variable into the primary register
int LoadVar(string Name, int Type)
{
	// cout << "# LoadVar: Promoting " << TypeToStr(Type);
	switch (Type)
	{
		case TYPE_BYTE:
			Type = TYPE_LONG;
			//cout << " to " << TypeToStr(Type) << endl;
			EmitLn("movsbl\t" + Name + ",%eax");
			break;
		case TYPE_UBYTE:
			Type = TYPE_LONG;
			//cout << " to " << TypeToStr(Type) << endl;
			//EmitLn("movl\t$0, %eax");
			EmitLn("movzbl\t" + Name + ",%eax");
			break;
		case TYPE_WORD:
			Type = TYPE_LONG;
			//cout << " to " << TypeToStr(Type) << endl;
			EmitLn("movswl\t" + Name + ",%eax");
			break;
		case TYPE_UWORD:
			Type = TYPE_LONG;
			//cout << " to " << TypeToStr(Type) << endl;
			EmitLn("movzwl\t" + Name + ",%eax");
			//EmitLn("andl\t$65535, %eax");
			break;
		case TYPE_LONG:
			Type = TYPE_LONG;
			//cout << " to " << TypeToStr(Type) << endl;
			EmitLn("movl\t" + Name + ", %eax");
			break;
		case TYPE_ULONG:
			Type = TYPE_ULONG;
			//cout << " to " << TypeToStr(Type) << endl;
			EmitLn("movl\t" + Name + ", %eax");
			break;
		default:
			Abort ("LoadVar: Unknown type");
			break;
	}
	return (Type);
}


// Store primary to variable
void StoreVar(string Name, int Type)
{
	switch (Type)
	{
		case TYPE_BYTE:
		case TYPE_UBYTE:
			EmitLn("mov\t%al, " + Name);
			break;
		case TYPE_WORD:
		case TYPE_UWORD:
			EmitLn("mov\t%ax, " + Name);
			break;
		case TYPE_LONG:
		case TYPE_ULONG:
			EmitLn("movl\t%eax, " + Name);
			break;
		default:
			Abort ("StoreVar: Unknown type");
			break;
	}
}


// Generate a move instruction sequence
void Move (int Type, string Source, string Dest)
{
	Abort("Move: ?");
	switch (Type)
	{
		case TYPE_BYTE:
			EmitLn("mov\t" + Source + ", %al");
			EmitLn("mov\t, %al" + Dest);
			break;
		case TYPE_WORD:
			EmitLn("mov\t" + Source + ", %ax");
			EmitLn("mov\t, %ax" + Dest);
			break;
		case TYPE_LONG:
			EmitLn("movl\t" + Source + ", %al");
			EmitLn("movl\t, %eax" + Dest);
			break;
		default:
			Abort ("Move: Unknown type");
			break;
	}		
}


// Convert a data item from one type to another
int Convert (int Type1, int Type2)
{
	int Type = TYPE_LONG;
	if ((Type1 == TYPE_ULONG) || (Type2 == TYPE_ULONG))
	{
		Type = TYPE_ULONG;
	}
	PrintTypes(Type1, Type2, Type);
	return (Type);
}

// Push primary register on to stack
void Push(int Type)
{
	// FIXME seems gcc always pushes longs !!!!!
	switch (Type)
	{
		case TYPE_BYTE:
		case TYPE_UBYTE:
		case TYPE_WORD:
		case TYPE_UWORD:
		case TYPE_LONG:
		case TYPE_ULONG:
			EmitLn("pushl\t%eax");
			break;
		default:
			Abort ("Push: Unknown type");
			break;
	}	
}
  

// Add top of stack to primary
int  PopAdd(int Type1, int Type2)
{
	int Type = Convert(Type1, Type2);
	EmitLn ("popl\t%ebx");
	EmitLn ("addl\t%ebx, %eax");
	return (Type);
}


// Subtract primary from top of stack
int PopSub(int Type1, int Type2)
{
	int Type = Convert(Type1, Type2);
	EmitLn ("popl\t%ebx");
	EmitLn ("subl\t%eax, %ebx");
	EmitLn ("mov\t%ebx, %eax");
	return (Type);
}


// Multiply top of stack by primary
int PopMul(int Type1, int Type2)
{
	int Type = Convert(Type1, Type2);
	EmitLn ("popl\t%ebx");
	EmitLn ("imull\t%ebx");
	return (Type);
}


bool isSignedType(int Type)
{
	int res;
	switch (Type)
	{
		case TYPE_BYTE:
		case TYPE_WORD:
		case TYPE_LONG:
			res = true;
			break;
		case TYPE_UBYTE:
		case TYPE_UWORD:
		case TYPE_ULONG:
			res = false;
			break;
		default:
			Abort ("isSignedType: Unknown type");
			break;
	}
	return (res);
}


// Divide top of stack by primary
int PopDiv(int Type1, int Type2)
{
	int Type = Convert(Type1, Type2);
	EmitLn ("movl\t%eax,%ebx");
	EmitLn ("popl\t%eax");	
	//if (isSignedType(Type1) && isSignedType(Type2))
	if (Type == TYPE_LONG)
	{
		EmitLn ("movl\t%eax, %edx");
		EmitLn ("sarl\t$31, %edx");
		EmitLn ("idivl\t%ebx");
	}
	else
	{
		EmitLn ("movl\t$0, %edx");	
		EmitLn ("divl\t%ebx");	
	}
	return (Type);
}


// AND top of stack with primary register
int PopAnd(int Type1, int Type2)
{
	int Type = Convert(Type1, Type2);
	EmitLn ("popl\t%ebx");
	EmitLn ("andl\t%ebx, %eax");
	return (Type);
}


// OR top of stack with primary register
int PopOr(int Type1, int Type2)
{
	int Type = Convert(Type1, Type2);
	EmitLn ("popl\t%ebx");
	EmitLn ("orl\t%ebx, %eax");
	return (Type);
}


// XOR top of stack with primary register
int PopXor(int Type1, int Type2)
{
	int Type = Convert(Type1, Type2);
	EmitLn ("popl\t%ebx");
	EmitLn ("xorl\t%ebx, %eax");
	return (Type);
}


// Compare top of stack with primary register
int PopCompare (int Type1, int Type2)
{
	int Type = Convert(Type1, Type2);
	EmitLn ("popl\t%ebx");
	EmitLn ("cmpl\t%eax, %ebx");		// Sets flags like subtract
	return (Type);
}


// Set primary TRUE if compare was = 
void SetEqual (int Type)
{
	EmitLn ("sete\t%al");			// No flag change
	//EmitLn ("movzx\t%al,%eax");		// No flag change
	EmitLn ("sall\t$31,%eax");		// Flags change
	EmitLn ("sarl\t$31,%eax");
}


// Set primary TRUE if compare was != 
void SetNotEqual(int Type)
{
	EmitLn ("setne\t%al");
	//EmitLn ("movzx\t%al,%eax");
	EmitLn ("sall\t$31,%eax");
	EmitLn ("sarl\t$31,%eax");
}


// Set primary TRUE if compare was > 
void SetGreater(int Type)
{
	if (isSignedType(Type))
	{
		EmitLn ("setg\t%al");
	}
	else
	{
		EmitLn ("seta\t%al");
	}
	//EmitLn ("movzx\t%al,%eax");
	EmitLn ("sall\t$31,%eax");
	EmitLn ("sarl\t$31,%eax");
}

// Set primary TRUE if compare was < 
void SetLess(int Type)
{
	if (isSignedType(Type))
	{
		EmitLn ("setl\t%al");
	}
	else
	{
		EmitLn ("setb\t%al");
	}
	//EmitLn ("movzx\t%al,%eax");
	EmitLn ("sall\t$31,%eax");
	EmitLn ("sarl\t$31,%eax");
}


// Set primary TRUE if compare was <= 
void SetLessOrEqual(int Type)
{
	if (isSignedType(Type))
	{
		EmitLn ("setle\t%al");
	}
	else
	{
		EmitLn ("setbe\t%al");
	}
	//EmitLn ("movzx\t%al,%eax");
	EmitLn ("sall\t$31,%eax");
	EmitLn ("sarl\t$31,%eax");
}


// Set primary TRUE if compare was >= 
void SetGreaterOrEqual(int Type)
{
	if (isSignedType(Type))
	{
		EmitLn ("setge\t%al");
	}
	else
	{
		EmitLn ("setae\t%al");
	}
	//EmitLn ("movzx\t%al,%eax");
	EmitLn ("sall\t$31,%eax");
	EmitLn ("sarl\t$31,%eax");
}


// Branch unconditional 
void Branch(string l)
{
	EmitLn("jmp\t" + l);
}


// Branch False 
void BranchFalse(string l)
{
	EmitLn("je\t" + l);
}


// Read variable into primary register
void ReadIt(string Name)
{
	EmitLn(string("call\t") + "READ");
	// FIXME Store(Name);
}


// Write variable from primary register
void WriteIt(int Type)
{
	//EmitLn("subl\t$12, %esp");
	EmitLn("pushl\t%eax");
	switch (Type)
	{
		case TYPE_BYTE:
			EmitLn(string("call\t") + "writesbyte");
			break;
		case TYPE_WORD:
			EmitLn(string("call\t") + "writesword");
			break;
		case TYPE_LONG:
			EmitLn(string("call\t") + "writeslong");
			break;
		case TYPE_UBYTE:
			EmitLn(string("call\t") + "writeubyte");
			break;
		case TYPE_UWORD:
			EmitLn(string("call\t") + "writeuword");
			break;
		case TYPE_ULONG:
			EmitLn(string("call\t") + "writeulong");
			break;
		default:
			Abort ("WriteIt: Unknown type");
		break;
	}
	//EmitLn("addl\t$16, %esp");
	EmitLn("addl\t$4, %esp");
}


// Write header info
void Header()
{
	EmitLn (".file\t\"fixme.c\"");
}


// Write the prolog
void Prolog()
{
	// This emits the prolog of C's main function
	// into which main program code is generated
	WriteLn(".globl main");
	EmitLn(".type\tmain, @function");
	PostLabel("main");
	EmitLn("pushl\t%ebp");
	EmitLn("movl\t%esp, %ebp");
	EmitLn("subl\t$8, %esp");
	EmitLn("andl\t$-16, %esp");
	EmitLn("movl\t$0, %eax");
	EmitLn("subl\t%eax, %esp");
}


// Write the epilog
void Epilog()
{
	// This emits the epilog of C's main function
	// and winds up the assembler junk
	EmitLn("movl\t$0, %eax");
	EmitLn("leave");
	EmitLn("ret");
	EmitLn(".size\tmain, .-main");
	EmitLn(".section\t.note.GNU-stack,\"\",@progbits");
	EmitLn(".ident\t\"GCC: (GNU) 3.3.2 20031218 (Gentoo Linux 3.3.2-r5, propolice-3.3-7)\"");
}


// Allocate storage for variable
void Allocate (string Name, int Type, string Val)
{
	static int startData = 1;

	WriteLn (string(".globl ") + Name);
	if (startData)
	{
		EmitLn (".data");
		startData = 0;
	}
	EmitLn (".align 4");
	EmitLn (".type\t" + Name + ", @object");
	switch (Type)
	{
		case TYPE_BYTE:
		case TYPE_UBYTE:
			EmitLn (".size\t" + Name + ", 1");
			PostLabel(Name);
			EmitLn (".byte\t" + Val);
			break;
		case TYPE_WORD:
		case TYPE_UWORD:
			EmitLn (".size\t" + Name + ", 2");
			PostLabel(Name);
			EmitLn (".value\t" + Val);
			break;
		case TYPE_LONG:
		case TYPE_ULONG:
			EmitLn (".size\t" + Name + ", 4");
			PostLabel(Name);
			EmitLn (".long\t" + Val);
			break;
		default:
			Abort("Alocate: unknown type");
			break;
	}
}


// Call a procedure
void Call(string Name)
{
     EmitLn("call\t" + Name);
}


// Load a parameter into the primary register
void LoadParameter(int ParNum, int Type)
{
	int offset;
	char offsetStr[100];
	offset = 4 + 4 + 4 * (Base - ParNum);
	sprintf(offsetStr, "%d", offset);
	switch (Type)
	{
		case TYPE_BYTE:
			EmitLn (string ("movsbl\t") + offsetStr + "(%ebp), %eax");
			break;
		case TYPE_UBYTE:
			EmitLn (string ("movzbl\t") + offsetStr + "(%ebp), %eax");
			break;
		case TYPE_WORD:
			EmitLn (string ("movswl\t") + offsetStr + "(%ebp), %eax");
			break;
		case TYPE_UWORD:
			EmitLn (string ("movzwl\t") + offsetStr + "(%ebp), %eax");
			break;
		case TYPE_LONG:
			EmitLn (string ("movl\t") + offsetStr + "(%ebp), %eax");
			break;
		case TYPE_ULONG:
			EmitLn (string ("movl\t") + offsetStr + "(%ebp), %eax");
			break;
		default:
			Abort ("LoadParameter: unknown type");
			break;
	}
}


// Store a parameter from the primary register
void StoreParam(int n)
{
	int offset;
	char offsetStr[100];
	offset = 4  + 4 + 4 * (Base - n);
	sprintf(offsetStr, "%d", offset);
	EmitLn (string("movl\t%eax, ") + offsetStr + "(%ebp)");
}

// Adjust the stack pointer upwards by N bytes
void CleanStack(int N)
{
	char Nstr[100];
	sprintf (Nstr, "%d", N);
	if (N > 0)
	{
		EmitLn (string ("addl\t$") + Nstr + ", %esp");
	}
}


// Write the prolog for a procedure
void ProcProlog(string Name, int k)
{
	WriteLn(string(".globl ") + Name);
	EmitLn (string(".type	") + Name + ",@function");
	PostLabel(Name);

	EmitLn ("pushl\t%ebp");
	EmitLn ("movl\t%esp,%ebp");

	char kStr[100];
	sprintf(kStr, "%d", k * 4);
	EmitLn (string ("subl\t$") + kStr + ", %esp");
}


// Write the epilog for a procedure
void ProcEpilog(string Name)
{
	EmitLn ("leave");
	EmitLn ("ret");
	EmitLn (string (".size	") + Name + ",.-" + Name);
}




